#!/usr/bin/perl
use warnings;
use strict;
use Carp;

use Test::More tests => 23;
use Test::Exception;

use POSIX qw( :signal_h WNOHANG );
use Time::HiRes qw( sleep );

use POWER::Utils;


sub FD_list {
    my $pid = shift;
    opendir my $fd, '/proc/'.$pid.'/fd' or croak "opendir: $!";
    return grep {!m/\A\.\.?\z/} readdir $fd;
};


my (@exit, $pipe_r, $pipe_w, $pipe2_r, $pipe2_w, $in, $out, @fd_list, $pid);
my ($CHLD_count, $CHLD_reaped) = (0, 0);

$SIG{CHLD} = sub {
    $CHLD_count++;
    while ((my $pid = waitpid(-1, &WNOHANG)) > 0) {
        $CHLD_reaped++;
    }
};

sub cb_read {
    close $pipe_w;
    $in = join q{}, <$pipe_r>;
    close $pipe_r;
}

@exit = System({ cmd => [q{bash -c 'exit 0'}], });
is_deeply(\@exit, [0,0,0],  'command exit 0');

@exit = System({ cmd => [q{bash -c 'exit 10'}], });
is_deeply(\@exit, [10,0,0], 'command exit 10');

@exit = System({ cmd => [q{kill -KILL $$}], });
is_deeply(\@exit, [0,9,0],  'command kill itself');

@exit = System({
    cmd         => ['sleep','3'],
    cb          => sub { kill 15, shift },
});
is_deeply(\@exit, [0,15,0], 'command killed by {cb} ({cb} got PID param)');

throws_ok { @exit=(); @exit = System({
    cmd         => ['sleep','3'],
    cb          => sub {
        $pid=shift;
        is(kill(0,$pid), 1, '  command now running');
        die "cb\n";
    },
}) } qr/cb/;
is_deeply(kill(0,$pid), 0,  '  command now killed because {cb} died');

@exit = System({
    cmd         => ['sleep 1; echo hi'],
    fh          => [],
    cb          => sub {
        sleep 0.2;    # to guarantee child is running when we do FD_list()
        @fd_list = FD_list($_[0]);
    },
});
is_deeply(\@fd_list, [],    'command executed with NO open filehandles');
is_deeply(\@exit, [1,0,0],  '  echo return exitcode 1');

ok($CHLD_count > 0,         'SIGCHLD happens some times');
is($CHLD_reaped, 0,         '  but no child was reaped');
exit if !fork;
sleep 0.2;  # avoid race with SIGCHLD
is($CHLD_reaped, 1,         '  now 1 child was reaped');


@exit = System({
    cmd         => ['sleep','2'],
    timeout     => 1,
});
is_deeply(\@exit, [0,14,0], 'command killed by SIGALRM because of {timeout}');

pipe $pipe_r, $pipe_w;
@exit = System({
    cmd         => ['fake_cmd','-c','sleep 1;echo -n $0'],
    real_cmd    => 'bash',
    fh          => [\*STDIN, $pipe_w],
    cb          => sub {
        sleep 0.2;    # to guarantee child is running when we do FD_list()
        @fd_list = FD_list($_[0]);
        &cb_read;
    },
});
is_deeply(\@exit, [0,0,0],  '{real_cmd} and STDOUT pipe working');
is_deeply($in, 'fake_cmd',  '  $0 is fake_cmd');
is_deeply(\@fd_list, [0,1], '  child has only STDIN and STDOUT');

pipe $pipe_r, $pipe_w;
pipe $pipe2_r, $pipe2_w;
@exit = System({
    cmd         => ['perl','-e','my ($n1,$n2) = <STDIN>; print $n1+$n2'],
    fh          => [$pipe_r, $pipe2_w],
    cb          => sub {
        close $pipe_r;
        close $pipe2_w;
        print {$pipe_w} "2\n5\n";
        close $pipe_w;
        $in = join q{}, <$pipe2_r>;
        close $pipe2_r;
    },
});
is_deeply(\@exit, [0,0,0],  'two pipes (STDIN&STDOUT) to calculator');
is_deeply($in, '7',         '  2+5=7');

sigprocmask(SIG_BLOCK, POSIX::SigSet->new(SIGINT), undef) or die $!;
$SIG{TERM} = 'IGNORE';
@exit = System({
    cmd                 => ['sleep','2'],
    timeout             => 1,
    cb                  => sub { kill SIGINT, shift },
});
is_deeply(\@exit, [0,2,0],  'child does not inherit masked SIGINT');
@exit = System({
    cmd                 => ['sleep','2'],
    timeout             => 1,
    cb                  => sub { kill SIGINT, $_[0]; kill SIGTERM, $_[0] },
    inherit_signals     => 1,
});
is_deeply(\@exit, [0,14,0],  'child inherit masked SIGINT and ignored SIGTERM');

$SIG{QUIT} = sub { die "got SIGQUIT\n" };
lives_ok { @exit=(); @exit = System({
    cmd                 => ['sleep','2'],
    timeout             => 1,
    cb                  => sub { kill SIGQUIT, $$; kill SIGQUIT, shift },
}) }                        'unable to kill parent with SIGQUIT';
is_deeply(\@exit, [0,3,0],  '  child killed by SIGQUIT');
throws_ok { System({
    cmd                 => ['sleep','2'],
    timeout             => 1,
    cb                  => sub { kill SIGQUIT, $$; kill SIGQUIT, shift },
    foreground          => 1,
}) } qr/got SIGQUIT/;

